import { OutgoingHttpHeaders } from 'node:http';
import { ObjectId } from 'bson';
export { EJSON, ObjectId } from 'bson';
import { Document, Filter, WithId, Sort, OptionalUnlessRequiredId, UpdateFilter, WithoutId } from 'mongodb';
export { Document, Filter, OptionalUnlessRequiredId, Sort, UpdateFilter, WithId, WithoutId } from 'mongodb';

type EmailPasswordAuthOptions = {
    email: string;
    password: string;
};
type ApiKeyAuthOptions = {
    apiKey: string;
};
type CustomJwtAuthOptions = {
    jwtTokenString: string;
};
type BearerTokenAuthOptions = {
    bearerToken: string;
};
type AuthOptions = EmailPasswordAuthOptions | ApiKeyAuthOptions | CustomJwtAuthOptions | BearerTokenAuthOptions;

/** Describes a Mongo Data API Error */
declare class MongoDataAPIError extends Error {
    /** The HTTP error code from the Mongo Data API */
    code: number;
    constructor(message: string, code: number);
}

type Nullable<T> = T | null;
/** A data API response object */
type DataAPIResponse<T> = {
    data?: T;
    error?: MongoDataAPIError;
};
type MongoClientConstructorOptions = {
    /**
     * Your atlas endpoint. Usually in the form of
     * https://data.mongodb-api.com/app/<clientAppId>/endpoint/data/v1
     * and retrievable from the Data API UI in Atlas. `clientAppId` is your Atlas App Service
     * friendly ID (not the hex string `_id` value).
     *
     * Your endpoint information is available at https://cloud.mongodb.com/v2/<projectId>#/dataAPI
     */
    endpoint: URL | string;
    /**
     * The [Atlas Data Source](https://cloud.mongodb.com/v2/<projectId>#/dataAPI). This tells
     * mongo which Cluster you want to run your queries against. The default data source
     * usually mimics your cluster's name, but you can find it in the Atlas UI.
     */
    dataSource: string;
    /**
     * An authentication method to use for your Atlas App Service. Supports API Key,
     * Email/Password, and Custom JWT as authentication methods.
     *
     * Read more: https://www.mongodb.com/docs/atlas/api/data-api/#authenticate-requests
     */
    auth: AuthOptions;
    /**
     * Provide a compatibility layer with a custom fetch implementation.
     */
    fetch?: typeof fetch;
    /**
     * Provide a custom Headers object for the request. In some situations such as testing,
     * this allows you to control the response from the mock server.
     */
    headers?: OutgoingHttpHeaders | Headers;
};
type ConnectionOptions = {
    dataSource: string;
    endpoint: string;
    headers: Record<string, string>;
};
type ConnectionOptionsWithDatabase = ConnectionOptions & {
    database: string;
};
type ConnectionOptionsWithCollection = ConnectionOptionsWithDatabase & {
    collection: string;
};
/**
 * Create a MongoDB-like client for communicating with the Atlas Data API
 *
 * ref: https://www.mongodb.com/docs/atlas/api/data-api-resources/
 *
 * The Data API client operates over https, which requires an available fetch()
 * implementation, either provided via polyfill, ponyfill, or natively available.
 *
 * To maximize compatibility, the client uses the application/ejson format in all
 * requests, converting mongo-style BSON to JSON and back again.
 */
declare class MongoClient {
    protected connection: ConnectionOptions;
    protected ftch: typeof fetch;
    constructor({ auth, dataSource, endpoint, fetch: customFetch, headers: h, }: MongoClientConstructorOptions);
    /** Select a database from within the data source */
    db(name: string): Database;
}
/**
 * Represents a Mongo Database object. Contains information about the selected
 * database, as well as connection information for performing
 * queries.
 */
declare class Database {
    protected connection: ConnectionOptionsWithDatabase;
    protected ftch: typeof fetch;
    constructor(name: string, connection: ConnectionOptions, customFetch?: typeof fetch);
    /**
     * Get a Collection from the Database
     * @param name The collection name
     * @typeParam `T` The type of document stored in the collection
     * @returns A Collection object of type `T`
     */
    collection<TSchema = Document>(name: string): Collection<TSchema>;
    /**
     * Change the fetch interface for this database. Returns a new Database object
     * @param customFetch A fetch interface to use for subsequent calls
     * @returns A new Database object with the custom fetch interface implemented
     */
    fetch(customFetch: typeof fetch): Database;
}
/**
 * Represents a Mongo Collection object of type `<T>`, providing query
 * methods in a fluent interface.
 */
declare class Collection<TSchema = Document> {
    protected connection: ConnectionOptionsWithCollection;
    protected ftch: typeof fetch;
    constructor(name: string, database: ConnectionOptionsWithDatabase, customFetch?: typeof fetch);
    /**
     * Change the fetch interface for this collection. Returns a new Collection object
     * @param customFetch A fetch interface to use for subsequent calls
     * @returns A new Collection object with the custom fetch interface implemented
     */
    fetch(customFetch: typeof fetch): Collection<Document>;
    /**
     * Find a single document in the collection. **note:** The `findOne()` operation does not provide a sort operation for finding a returning the first record in a set. If you must `findOne` and `sort`, use the `find()` operation instead.
     * [example](https://www.mongodb.com/docs/atlas/api/data-api-resources/#find-a-single-document)
     *
     * @param filter A [MongoDB Query Filter](https://www.mongodb.com/docs/manual/tutorial/query-documents/). The findOne action returns the first document in the collection that matches this filter.
     *
     * If you do not specify a filter, the action matches all document in the collection.
     * @param options Query options
     * @param options.projection A [MongoDB Query Projection](https://www.mongodb.com/docs/manual/tutorial/project-fields-from-query-results/). Depending on the projection, the returned document will either omit specific fields or include only specified fields or values.
     * @returns A matching document, or `null` if no document is found
     */
    findOne(filter?: Filter<TSchema>, options?: {
        projection?: Document;
        sort?: Document;
    }): Promise<DataAPIResponse<Nullable<WithId<TSchema>>>>;
    /**
     * Find multiple documents in the collection matching a query.
     * [example](https://www.mongodb.com/docs/atlas/api/data-api-resources/#find-multiple-documents)
     *
     * @param filter A [MongoDB Query Filter](https://www.mongodb.com/docs/manual/tutorial/query-documents/). The find action returns documents in the collection that match this filter.
     *
     * If you do not specify a `filter`, the action matches all document in the collection.
     *
     * If the filter matches more documents than the specified `limit`, the action only returns a subset of them. You can use `skip` in subsequent queries to return later documents in the result set.
     * @param options Query options
     * @param options.projection A [MongoDB Query Projection](https://www.mongodb.com/docs/manual/tutorial/project-fields-from-query-results/). Depending on the projection, the returned document will either omit specific fields or include only specified fields or values.
     * @param options.sort A [MongoDB Sort Expression](https://www.mongodb.com/docs/manual/reference/operator/aggregation/sort/). Matched documents are returned in ascending or descending order of the fields specified in the expression.
     * @param options.limit The maximum number of matched documents to include in the returned result set. Each request may return up to 50,000 documents.
     * @param options.skip The number of matched documents to skip before adding matched documents to the result set.
     * @returns An array of matching documents
     */
    find(filter?: Filter<TSchema>, options?: {
        projection?: Document;
        sort?: Sort;
        limit?: number;
        skip?: number;
    }): Promise<DataAPIResponse<WithId<TSchema>[]>>;
    /**
     * Insert a single document into the collection. [read more](https://www.mongodb.com/docs/atlas/api/data-api-resources/#insert-a-single-document)
     * @param document The document to insert (of type `T`)
     * @returns The insertOne action returns the _id value of the inserted document as a string in the insertedId field
     */
    insertOne(document: OptionalUnlessRequiredId<TSchema>): Promise<DataAPIResponse<{
        insertedId: ObjectId;
    }>>;
    /**
     * Insert several documents into the collection. [read more](https://www.mongodb.com/docs/atlas/api/data-api-resources/#insert-multiple-documents)
     * @param docs An array of documents to insert (of type `T[]`)
     * @returns The insertMany action returns the _id values of all inserted documents as an array of strings in the insertedIds field
     */
    insertMany(documents: OptionalUnlessRequiredId<TSchema>[]): Promise<DataAPIResponse<{
        insertedIds: string[];
    }>>;
    /**
     * Update a single matching document in the collection. [read more](https://www.mongodb.com/docs/atlas/api/data-api-resources/#update-a-single-document)
     * @param filter A [MongoDB Query Filter](https://www.mongodb.com/docs/manual/tutorial/query-documents/). The updateOne action modifies the first document in the collection that matches this filter.
     * @param update A [MongoDB Update Expression](https://www.mongodb.com/docs/manual/tutorial/update-documents/) that specifies how to modify the matched document.
     * @param options Query options
     * @param options.upsert The upsert flag only applies if no documents match the specified filter. If true, the updateOne action inserts a new document that matches the filter with the specified update applied to it.
     * @returns The updateOne action returns:
     *   - the number of documents that the filter matched in the `matchedCount` field
     *   - the number of matching documents that were updated in the `modifiedCount` field
     *
     * If upsert is set to true and no documents match the filter, the action returns the `_id` value of the inserted document as a string in the `upsertedId` field
     */
    updateOne(filter: Filter<TSchema>, update: UpdateFilter<TSchema> | Partial<TSchema>, options?: {
        upsert?: boolean;
    }): Promise<DataAPIResponse<{
        matchedCount: number;
        modifiedCount: number;
        upsertedId?: string;
    }>>;
    /**
     * Update multiple documents in the collection matching a criteria. [read more](https://www.mongodb.com/docs/atlas/api/data-api-resources/#update-multiple-documents)
     * @param filter A [MongoDB Query Filter](https://www.mongodb.com/docs/manual/tutorial/query-documents/). The updateMany action modifies all documents in the collection that match this filter.
     * @param update A [MongoDB Update Expression](https://www.mongodb.com/docs/manual/tutorial/update-documents/) that specifies how to modify matched documents.
     * @param options Query options
     * @param options.upsert The upsert flag only applies if no documents match the specified filter. If true, the updateMany action inserts a new document that matches the filter with the specified update applied to it.
     * @returns The updateMany action returns:
     *   - the number of documents that the filter matched in the `matchedCount` field
     *   - the number of matching documents that were updated in the `modifiedCount` field
     *
     * If upsert is set to true and no documents match the filter, the action returns the `_id` value of the inserted document as a string in the `upsertedId` field
     */
    updateMany(filter: Filter<TSchema>, update: UpdateFilter<TSchema>, { upsert }?: {
        upsert?: boolean;
    }): Promise<DataAPIResponse<{
        matchedCount: number;
        modifiedCount: number;
        upsertedId?: string;
    }>>;
    /**
     * Replace a single document in the collection. [read more](https://www.mongodb.com/docs/atlas/api/data-api-resources/#replace-a-single-document)
     * @param filter A [MongoDB Query Filter](https://www.mongodb.com/docs/manual/tutorial/query-documents/). The replaceOne action overwrites the first document in the collection that matches this filter.
     * @param replacement The replacement document.
     * @param options Query options
     * @param options.upsert The upsert flag only applies if no documents match the specified filter. If true, the replaceOne action inserts a new document that matches the filter with the specified replacement applied to it.
     * @returns The replaceOne action returns:
     *   - the number of documents that the filter matched in the `matchedCount` field
     *   - the number of matching documents that were updated in the `modifiedCount` field
     *
     * If upsert is set to true and no documents match the filter, the action returns the `_id` value of the inserted document as a string in the `upsertedId` field
     */
    replaceOne(filter: Filter<TSchema>, replacement: WithoutId<TSchema>, options?: {
        upsert?: boolean;
    }): Promise<DataAPIResponse<{
        matchedCount: number;
        modifiedCount: number;
        upsertedId?: string;
    }>>;
    /**
     * Delete a single document in the collection. [read more](https://www.mongodb.com/docs/atlas/api/data-api-resources/#delete-a-single-document)
     * @param filter A [MongoDB Query Filter](https://www.mongodb.com/docs/manual/tutorial/query-documents/). The deleteOne action deletes the first document in the collection that matches this filter.
     * @returns The deleteOne action returns the number of deleted documents in the deletedCount field
     */
    deleteOne(filter: Filter<TSchema>): Promise<DataAPIResponse<{
        deletedCount: number;
    }>>;
    /**
     * Delete multiple documents from the collection matching a filter. [read more](https://www.mongodb.com/docs/atlas/api/data-api-resources/#delete-multiple-documents)
     * @param filter A [MongoDB Query Filter](https://www.mongodb.com/docs/manual/tutorial/query-documents/). The deleteMany action deletes all documents in the collection that match this filter.
     * @returns The deleteMany action returns the number of deleted documents in the deletedCount field
     */
    deleteMany(filter: Filter<TSchema>): Promise<DataAPIResponse<{
        deletedCount: number;
    }>>;
    /**
     * Run an aggregation pipeline on the collection. [read more](https://www.mongodb.com/docs/atlas/api/data-api-resources/#aggregate-documents)
     * @param pipeline A [MongoDB Aggregation Pipeline](https://www.mongodb.com/docs/manual/core/aggregation-pipeline/).
     * @returns The aggregate action returns the result set of the final stage of the pipeline as an array of documents, automatically unwrapping the Atlas `documents` field
     */
    aggregate<TOutput = Document>(pipeline: Document[]): Promise<DataAPIResponse<TOutput[]>>;
    /**
     * Call a raw Data API resource
     * @param method The Data API resource to call
     * @param body The JSON body to include, merged with the collection, dataSource, and database names
     * @returns The response from the Data API, or a DataAPIError if the request failed
     */
    callApi<T = unknown>(method: string, body: Record<string, unknown>): Promise<DataAPIResponse<T>>;
}

export { ApiKeyAuthOptions, AuthOptions, BearerTokenAuthOptions, Collection, CustomJwtAuthOptions, Database, EmailPasswordAuthOptions, MongoClient, MongoClientConstructorOptions, MongoDataAPIError };
